from .buddhabrot import buddhabrot
from .common import Canvas, Result
from .images import images
from .iterfuncs import funcs
from .julia import julia
from .lyapunov import lyapunov
from .mandelbrot import mandelbrot
from .randomwalk import construct_moves, randomwalk

__all__ = [
    "buddhabrot",
    "images",
    "julia",
    "julia_series",
    "lyapunov",
    "mandelbrot",
    "randomwalk",
    "construct_moves",
    "Result",
    "Canvas",
    "funcs",
]
__version__ = "0.1.0"



import itertools as itt

import numpy as np
from numba import jit
from numpy import array
from numpy.random import randint

from pyfracgen.common import Canvas3D, Result
from pyfracgen.types import Lattice3D, Moves3D


def construct_moves(basis: Moves3D) -> Moves3D:

    basis = np.r_[basis, -1 * basis, [array([0, 0, 0])]]
    nonnull = list(
        itt.filterfalse(
            lambda x: not np.any(x), (b0 + b1 for b0, b1 in itt.combinations(basis, 2))
        )
    )
    moves: Moves3D = np.unique(nonnull, axis=0)
    return moves


@jit  # type: ignore[misc]
def _randomwalk_paint(
    lattice: Lattice3D,
    moves: Moves3D,
    niter: int,
) -> None:

    nmoves = len(moves)
    h, w, d = lattice.shape
    indices = array([h, w, d]) / 2.0
    for iteration in range(niter):
        move = moves[randint(0, nmoves)]
        indices += move
        iy, ix, iz = int(indices[0] % h), int(indices[1] % w), int(indices[2] % d)
        lattice[iy, ix, iz] = iteration

    lattice /= np.amax(lattice)


class RandomWalk(Canvas3D):
    def paint(self, moves: Moves3D, niter: int) -> None:
        _randomwalk_paint(self.lattice, moves, niter)


def randomwalk(
    moves: Moves3D,
    niter: int,
    width: int = 5,
    height: int = 4,
    depth: int = 1,
    dpi: int = 300,
) -> Result:

    canvas = RandomWalk(width, height, depth, dpi)
    canvas.paint(moves, niter)
    return canvas.result



"""Utility objects and methods used across the library."""
from __future__ import annotations

import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np

from pyfracgen.types import Bound, Lattice, Lattice3D

RESULT_DEFAULT_SAVE = Path("save.pickle")


@dataclass(frozen=True)
class Result:

    image_array: Lattice
    width_inches: int
    height_inches: int
    dpi: int

    @classmethod
    def load(cls, file: Path) -> Result:
        with open(file, "rb") as f:
            res = pickle.load(f)
        return cls(*res)

    def save(self, name: Path = RESULT_DEFAULT_SAVE) -> None:
        res = [self.image_array, self.width_inches, self.height_inches, self.dpi]
        with open(name, "wb") as f:
            pickle.dump(res, f)


class Canvas:
    def __init__(self, width: int, height: int, dpi: int):
        self.lattice: Lattice = np.zeros((height * dpi, width * dpi), dtype=np.float64)
        self.width = width
        self.height = height
        self.dpi = dpi

    @property
    def result(self) -> Result:
        return Result(self.lattice, self.width, self.height, self.dpi)

    def paint(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError


class Canvas3D(Canvas):
    def __init__(self, width: int, height: int, depth: int, dpi: int) -> None:
        super().__init__(width, height, dpi)
        self.lattice: Lattice3D = np.dstack(
            [np.zeros(self.lattice.shape) for _ in range(depth)]
        )


class CanvasBounded(Canvas):
    def __init__(
        self, width: int, height: int, dpi: int, xbound: Bound, ybound: Bound
    ) -> None:
        super().__init__(width, height, dpi)
        ny, nx = self.lattice.shape
        self.xbound = xbound
        self.ybound = ybound
        (xmin, xmax), (ymin, ymax) = xbound, ybound
        self.xvals = np.array(
            [xmin + i * (xmax - xmin) / nx for i in range(nx)], dtype=np.float64
        )
        self.yvals = np.array(
            [ymin + i * (ymax - ymin) / ny for i in range(ny)], dtype=np.float64
        )

    @property
    def bounds(self) -> tuple[Bound, Bound]:
        return (self.xbound, self.ybound)



import math
from typing import Iterable, Iterator

import numpy as np
from numba import jit
from numpy.random import random

from pyfracgen.common import CanvasBounded, Result
from pyfracgen.iterfuncs.funcs import power
from pyfracgen.mandelbrot import mandelbrot
from pyfracgen.types import Bound, Boxes, ComplexSequence, IterFunc, Lattice


@jit  # type: ignore[misc]
def threshold_round_array(arr: Lattice, threshold: float = 0.5) -> None:

    w, h = arr.shape
    for iy in range(w):
        for ix in range(h):
            current = arr[iy, ix]
            if current - int(current) < threshold:
                rounded = math.floor(current)
            else:
                rounded = math.ceil(current)
            arr[iy, ix] = rounded


@jit  # type: ignore[misc]
def round_array_preserving_sum(arr: Lattice) -> None:

    target = arr.sum()
    best = (math.inf, 0.5)
    # Simply checking 100 values between 0 and 1 works well enough
    for thresh in np.linspace(0, 1, 100):
        check = arr.copy()
        threshold_round_array(check, thresh)
        dist = abs(check.sum() - target)
        if dist < best[0]:
            best = (dist, thresh)
    threshold_round_array(arr, best[-1])


@jit  # type: ignore[misc]
def _compute_cvals(
    ncvals: int,
    bounds: tuple[Bound, Bound],
    boxes: tuple[Boxes, Boxes],
    energy_grid: Lattice,
    random_fraction: float,
) -> ComplexSequence:

    nr = round(ncvals * random_fraction)
    cvals = []
    (xmin, xmax), (ymin, ymax) = bounds
    # Randomly sampled starting points
    for _ in range(nr):
        c = xmin + (random() * (xmax - xmin)) + 1j * (ymin + (random() * (ymax - ymin)))
        cvals.append(c)

    # Energy grid sampled starting points
    if random_fraction < 1.0:
        ni = round(ncvals * (1 - random_fraction))
        energy_grid = (energy_grid / energy_grid.sum()) * ni
        round_array_preserving_sum(energy_grid)

        xboxes, yboxes = boxes
        for iy, (ylo, yhi) in enumerate(yboxes):
            for ix, (xlo, xhi) in enumerate(xboxes):
                nsamples = int(energy_grid[iy, ix])
                xadd = xlo + (random(nsamples) * (xhi - xlo))
                yadd = ylo + (random(nsamples) * (yhi - ylo))
                cvals.extend(xadd + 1j * yadd)

    return np.array(cvals)


@jit  # type: ignore[misc]
def _buddhabrot_paint(
    bounds: tuple[Bound, Bound],
    lattice: Lattice,
    cvals: ComplexSequence,
    update_func: IterFunc,
    maxiter: int,
    horizon: float,
) -> None:

    (xmin, xmax), (ymin, ymax) = bounds
    height, width = lattice.shape
    for c in cvals:
        z = c
        trial_sequence = []
        sequence = []
        for _ in range(maxiter):
            az = np.abs(z)
            trial_sequence.append(z)
            if az > horizon:
                sequence.extend(trial_sequence)
                break
            z = update_func(z, c)
        for c in sequence:
            indx = int((c.real - xmin) / (xmax - xmin) * width)
            indy = int((c.imag - ymin) / (ymax - ymin) * height)
            if (indx < 0 or indx >= width) or (indy < 0 or indy >= height):
                continue
            lattice[indy, indx] += 1


class Buddhabrot(CanvasBounded):
    @property
    def boxes(self) -> tuple[Boxes, Boxes]:

        xboxes = [
            (self.xvals[ix], self.xvals[ix + 1]) for ix in range(len(self.xvals) - 1)
        ]
        yboxes = [
            (self.yvals[ix], self.yvals[ix + 1]) for ix in range(len(self.yvals) - 1)
        ]
        *_, (_, lastx) = xboxes
        *_, (_, lasty) = yboxes
        (_, xmax), (_, ymax) = self.bounds
        xboxes += [(lastx, xmax)]
        yboxes += [(lasty, ymax)]
        return np.array(xboxes), np.array(yboxes)

    def compute_cvals(
        self,
        ncvals: int,
        energy_grid: Lattice,
        random_fraction: float = 0.25,
    ) -> ComplexSequence:
        cvals: ComplexSequence = _compute_cvals(
            ncvals,
            self.bounds,
            self.boxes,
            energy_grid,
            random_fraction,
        )
        return cvals

    def paint(
        self,
        cvals: ComplexSequence,
        update_func: IterFunc,
        maxiter: int,
        horizon: float,
    ) -> None:

        _buddhabrot_paint(
            self.bounds,
            self.lattice,
            cvals,
            update_func,
            maxiter,
            horizon,
        )


def buddhabrot(
    xbound: Bound,
    ybound: Bound,
    ncvals: int,
    update_func: IterFunc = power,
    width: int = 5,
    height: int = 4,
    dpi: int = 300,
    maxiters: Iterable[int] = (100,),
    horizon: float = 1.0e6,
    random_fraction: float = 0.25,
) -> Iterator[Result]:

    mdbres = mandelbrot(
        xbound=xbound,
        ybound=ybound,
        update_func=update_func,
        width=width,
        height=height,
        dpi=dpi,
        maxiter=500,
        horizon=2.5,
        log_smooth=False,
    )
    canvi = [(m, Buddhabrot(width, height, dpi, xbound, ybound)) for m in maxiters]
    (_, first), *_ = canvi
    cvals = first.compute_cvals(
        ncvals, mdbres.image_array, random_fraction=random_fraction
    )
    for maxiter, canvas in canvi:
        canvas.paint(cvals, update_func, maxiter, horizon)
        yield canvas.result



from pathlib import Path
from typing import Sequence

import matplotlib.animation as animation
import numpy as np
from matplotlib import colormaps, colors
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure, figaspect

from pyfracgen.common import Result

ANIMATION_DEFAULT_SAVE = Path("ani.gif")
DEFAULT_COLORMAP = colormaps["hot"]


def get_stacked_cmap(cmap: colors.Colormap, nstacks: int) -> colors.Colormap:

    cline = np.vstack([np.array(cmap(np.linspace(0, 1, 200))) for _ in range(nstacks)])
    mymap = colors.LinearSegmentedColormap.from_list("stacked", cline)
    return mymap


def image(
    res: Result,
    cmap: colors.Colormap = DEFAULT_COLORMAP,
    ticks: bool = False,
    gamma: float = 0.3,
    vert_exag: float = 0.0,
    ls: tuple[int, int] = (315, 10),
) -> tuple[Figure, Axes]:

    fig, ax0 = plt.subplots(figsize=figaspect(res.image_array), dpi=res.dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    norm = colors.PowerNorm(gamma)
    if vert_exag > 0.0:
        light = colors.LightSource(azdeg=ls[0], altdeg=ls[1])
        shade = light.shade(
            res.image_array, cmap=cmap, norm=norm, vert_exag=vert_exag, blend_mode="hsv"
        )
        ax0.imshow(shade, origin="lower")
    else:
        ax0.imshow(res.image_array, origin="lower", cmap=cmap, norm=norm)
    if not ticks:
        plt.axis("off")
    fs = plt.gcf()
    fs.set_size_inches(res.width_inches, res.height_inches)
    return fig, ax0


def nebula_image(
    results: tuple[Result, Result, Result],
    ticks: bool = False,
    gamma: float = 1.0,
) -> tuple[Figure, Axes]:

    blue, green, red = results
    fig, ax0 = plt.subplots(figsize=figaspect(blue.image_array), dpi=blue.dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    arrays = [red.image_array, green.image_array, blue.image_array]
    final = np.dstack([arr / np.amax(arr) for arr in arrays])
    ax0.imshow(final**gamma, origin="lower")
    if not ticks:
        plt.axis("off")
    fs = plt.gcf()
    fs.set_size_inches(blue.width_inches, blue.height_inches)
    return fig, ax0


def markus_lyapunov_image(
    res: Result,
    cmap_neg: colors.Colormap,
    cmap_pos: colors.Colormap,
    gammas: tuple[float, float] = (1.0, 1.0),
    ticks: bool = False,
) -> tuple[Figure, Axes]:

    fig, ax0 = plt.subplots(figsize=figaspect(res.image_array), dpi=res.dpi)
    ax0.imshow(
        np.ma.masked_where(  # type: ignore[no-untyped-call]
            res.image_array > 0.0, res.image_array
        ),
        cmap=cmap_neg,
        origin="lower",
        norm=colors.PowerNorm(gammas[0]),
    )
    ax0.imshow(
        np.ma.masked_where(  # type: ignore[no-untyped-call]
            res.image_array < 0.0, res.image_array
        ),
        cmap=cmap_pos,
        origin="lower",
        norm=colors.PowerNorm(gammas[1]),
    )
    fig.subplots_adjust(0, 0, 1, 1)
    if not ticks:
        plt.axis(ticks)
    fs = plt.gcf()
    fs.set_size_inches(res.width_inches, res.height_inches)
    return fig, ax0


def randomwalk_image(
    res: Result,
    cmap: colors.Colormap = DEFAULT_COLORMAP,
    ticks: bool = False,
    gamma: float = 0.3,
    alpha_scale: float = 1.0,
) -> tuple[Figure, Axes]:

    fig, ax0 = plt.subplots(figsize=figaspect(res.image_array[:, :, 0]), dpi=res.dpi)
    max_ind = float(res.image_array.shape[-1] + 1)
    for i in range(res.image_array.shape[-1]):
        im = res.image_array[..., i]
        im = np.ma.masked_where(im == 0, im)  # type: ignore[no-untyped-call]
        alpha = alpha_scale * (1 - (i + 1) / max_ind)
        norm = colors.PowerNorm(gamma)
        ax0.imshow(
            im, origin="lower", alpha=alpha, cmap=cmap, norm=norm, interpolation=None
        )
    fig.subplots_adjust(0, 0, 1, 1)
    if not ticks:
        plt.axis(ticks)
    fs = plt.gcf()
    fs.set_size_inches(res.width_inches, res.height_inches)
    return fig, ax0


def save_animation(
    series: Sequence[Result],
    cmap: colors.Colormap = DEFAULT_COLORMAP,
    fps: int = 15,
    bitrate: int = 1800,
    file: Path = ANIMATION_DEFAULT_SAVE,
    ticks: bool = True,
    gamma: float = 0.3,
    vert_exag: float = 0,
    ls: tuple[int, int] = (315, 10),
) -> None:

    fig = plt.figure()
    fig.subplots_adjust(0, 0, 1, 1)
    if not ticks:
        plt.axis(ticks)
    fs = plt.gcf()
    norm = colors.PowerNorm(gamma)
    light = colors.LightSource(azdeg=ls[0], altdeg=ls[1])
    first, *_ = series
    fs.set_size_inches(first.width_inches, first.height_inches)
    writer = animation.PillowWriter(
        fps=fps, metadata=dict(artist="Me"), bitrate=bitrate
    )

    ims = []
    for res in series:
        shade = light.shade(
            res.image_array, cmap=cmap, norm=norm, vert_exag=vert_exag, blend_mode="hsv"
        )
        im = plt.imshow(shade, origin="lower", norm=norm)
        ims.append([im])

    if file.suffix != ".gif":
        file = file.with_suffix(".gif")
    ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True, repeat_delay=1000)
    ani.save(file, dpi=first.dpi, writer=writer)
